#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  7 09:29:29 2021

@author: mike
"""
import numpy as np
import xarray as xr
import pandas as pd
import os
import glob
from tethys_utils.processing import write_pkl_zstd, process_datasets, prepare_results, assign_station_id, make_run_date_key
from tethys_utils.s3 import process_run_date, update_results_s3, put_remote_dataset, put_remote_agg_stations, put_remote_agg_datasets, s3_connection
from shapely.geometry import shape, mapping, Point, box
import copy
import rasterio
import concurrent.futures
from tethys_utils.grid import Grid


###########################################
### Parameters


############################################
### Functions


def parse_images(path_str):
    """

    """
    f2 = glob.glob(path_str)
    f3 = {f: os.path.getsize(f) for f in f2}
    max1 = max(f3.values())
    max_f = [f for f in f3 if f3[f] == max1][0]

    return f3, max_f


def process_image(image, parameter, x_name='x', y_name='y', band=1):
    """

    """
    xr1 = xr.open_rasterio(image)
    xr1 = xr1.rename({x_name: 'lon', y_name: 'lat'}).sel(band=band).drop('band')
    xr1.name = parameter

    return xr1


############################################
### Class


class Raster(object):
    """

    """
    ## Initial import and assignment function
    def __init__(self, path_str, time, height, x_name='x', y_name='y', band=1, dataset_list=None, remote=None, processing_code=None, public_url=None, run_date=None):
        f_dict, max_f = parse_images(path_str)

        # Run checks
        src = rasterio.open(max_f)
        crs = src.crs.to_epsg()

        if crs != 4326:
            raise ValueError('Raster CRS is in epsg: ' + str(crs) + ', but should be 4326')

        src.close()

        # Set attrs
        setattr(self, 'x_name', x_name)
        setattr(self, 'y_name', y_name)
        setattr(self, 'max_image', max_f)
        setattr(self, 'images', f_dict)
        setattr(self, 'band', band)
        setattr(self, 'time', time)
        setattr(self, 'height', height)

        if isinstance(dataset_list, list):
            grid = Grid(dataset_list, remote, processing_code, public_url, run_date)
            setattr(self, 'update_aggregates', self.grid.update_aggregates)

        else:
            grid = Grid()

        setattr(self, 'grid', grid)

        pass


    def open_big_one(self):
        """

        """
        xr1 = xr.open_rasterio(self.max_image)

        return xr1


    def determine_grid_block_size(self, starting_x_size=100, starting_y_size=100, increment=100, min_size=800, max_size=1100):
        """

        """
        parameter = self.grid.datasets[0]['parameter']
        xr1 = process_image(self.max_image, parameter, x_name=self.x_name, y_name=self.y_name, band=1)
        self.grid.load_data(xr1.to_dataset(), parameter, self.time, self.height)
        size_dict = self.grid.determine_grid_block_size(starting_x_size, starting_y_size, increment, min_size, max_size)

        setattr(self, 'grid_size_dict', size_dict)

        res = xr1.attrs['res'][0]
        setattr(self, 'grid_res', res)

        return size_dict


    def save_results(self, x_size, y_size, threads=30):
        """

        """
        ## Iterate through the images
        images = list(self.images.keys())
        images.sort()

        for tif in images:
            print(tif)

            parameter = self.grid.datasets[0]['parameter']
            xr1 = process_image(self.max_image, parameter, x_name=self.x_name, y_name=self.y_name, band=1)
            self.grid.load_data(xr1.to_dataset(), parameter, self.time, self.height)

            self.grid.save_results(x_size, y_size, threads=threads)













#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May  7 09:29:29 2021

@author: mike
"""
import numpy as np
import xarray as xr
import pandas as pd
from tethys_utils.processing import write_pkl_zstd, prepare_results, assign_ds_ids
from tethys_utils.s3 import process_run_date, update_results_s3, put_remote_dataset, put_remote_agg_stations, put_remote_agg_datasets, s3_connection
# from shapely.geometry import shape, mapping, Point, box
# import copy
# import rasterio
import concurrent.futures


###########################################
### Parameters


############################################
### Functions


def _split_grid(arr, x_size, y_size, x_name='lon', y_name='lat'):
    """
    Function to split an n-dimensional dataset along the x and y dimensions. Optionally, add time and height dimensions if the array does not aready contain them.

    Parameters
    ----------
    arr : DataArray
        An xarray DataArray with at least x and y dimensions. It can have any number of dimensions, though it probably does not make much sense to have greater than 4 dimensions.
    x_size : int
        The size or length of the smaller grids in the x dimension.
    y_size : int
        The size or length of the smaller grids in the y dimension.
    x_name : str
        The x dimension name.
    y_name : str
        The y dimension name.

    Returns
    -------
    List of DataArrays
        The result contains none of the original attributes.
    """
    ## Get the dimension data
    dims = arr.dims

    # Get other array info
    x_index = dims.index(x_name)
    y_index = dims.index(y_name)
    data_name = arr.name

    arr_shape = arr.shape

    m = arr_shape[x_index]
    n = arr_shape[y_index]
    dtype = arr.dtype

    ## Build the new regular array to be queried
    y_diff = arr[y_name].diff(y_name, 1).median().values
    x_diff = arr[x_name].diff(x_name, 1).median().values

    bpx = ((m-1)//x_size + 1) # blocks per x
    bpy = ((n-1)//y_size + 1) # blocks per y
    M = x_size * bpx
    N = y_size * bpy

    x_y = list(arr_shape)
    x_y[x_index] = M
    x_y[y_index] = N

    sel1 = tuple(slice(0, s) for s in arr_shape)

    A = np.nan * np.ones(x_y)
    A[sel1] = arr

    # x array
    x_start = arr[x_name][0].values
    x_int = M * x_diff
    x_end = x_start + x_int
    xs = np.arange(x_start, x_end, x_diff)

    # y array
    y_start = arr[y_name][0].values
    y_int = N * y_diff
    y_end = y_start + y_int
    ys = np.arange(y_start, y_end, y_diff)

    # Coords
    coords = []
    new_dims = []
    for d in dims:
        name = d
        if d == x_name:
            c = xs
        elif d == y_name:
            c = ys
        else:
            c = arr[d]
        coords.extend([c])
        new_dims.extend([name])

    # New DataArray
    A1 = xr.DataArray(A, coords=coords, dims=new_dims, name=data_name)

    block_list = []
    previous_x = 0
    for x_block in range(bpy):
        previous_x = x_block * x_size
        previous_y = 0
        for y_block in range(bpx):
            previous_y = y_block * y_size
            x_slice = slice(previous_x, previous_x+x_size)
            y_slice = slice(previous_y, previous_y+y_size)

            sel2 = list(sel1)
            sel2[x_index] = x_slice
            sel2[y_index] = y_slice

            block = A1[tuple(sel2)]

            # remove nans
            block = block.dropna(y_name, 'all')
            block = block.dropna(x_name, 'all')

            ## append
            if block.size:
                block_list.append(block.astype(dtype))

    return block_list


############################################
### Class


class Grid(object):
    """

    """
    ## Initial import and assignment function
    def __init__(self, dataset_list=None, remote=None, processing_code=None, public_url=None, run_date=None):
        """

        """
        if isinstance(dataset_list, list):
            self.process_datasets(dataset_list, remote, processing_code, public_url, run_date)

        pass


    def process_datasets(self, dataset_list, remote, processing_code, public_url=None, run_date=None):
        """

        """
        ### Create dataset_ids
        datasets = assign_ds_ids(dataset_list)

        ## Checks
        grid_bool = all([d['spatial_distribution'] == 'grid' for d in datasets])
        if not grid_bool:
            raise ValueError('All values of spatial_distribution in the datasets should be grid.')

        grouping_len = len(np.unique([d['grouping'] for d in datasets]))
        if grouping_len > 1:
            raise ValueError('All values of grouping in the datasets must be the same.')

        grouping = datasets[0]['grouping']

        ### Determine the last run date and process the old data if old enough
        run_date_dict = process_run_date(processing_code, datasets, remote, run_date)
        max_run_date_key = max(list(run_date_dict.values()))

        ## Create the data_dict
        # data_dict = {d['dataset_id']: [] for d in datasets}

        ## Save objects
        setattr(self, 's3_remote', remote)
        setattr(self, 'public_url', public_url)
        setattr(self, 'datasets', datasets)
        setattr(self, 'run_date_dict', run_date_dict)
        setattr(self, 'max_run_date_key', max_run_date_key)
        setattr(self, 'processing_code', processing_code)
        setattr(self, 'grouping', grouping)

        pass


    def load_data(self, data, parameter, time=None, height=None):
        """
        Load an xr.Dataset into the class and perform checks.

        Parameters
        ----------
        data : xr.Dataset
            The data
        parameter : str
            The parameter name.
        time : str, pd.Timestamp, or None
            If time is not already in the data as a dimension (if this is a raster), then this values will be added to the data.
        height : int, float, or None
            If height is not already in the data as a dimension, then this values will be added to the data.

        """
        if hasattr(self, 'data'):
            delattr(self, 'data')

        arr = data.copy()

        ## Get the dimension data
        dims = arr.dims

        # Check and add dimensions if necessary
        if not 'lon' in dims:
            raise ValueError('lon must be a dimension.')
        if not 'lat' in dims:
            raise ValueError('lat must be a dimension.')

        if not 'time' in dims:
            if not isinstance(time, (str, pd.Timestamp)):
                raise ValueError("If time is not part of the array's dimensions, then a string or Timestamp must be passed to time.")
            else:
                print('time will be added to the data.')
                time1 = pd.Timestamp(time)
                arr = arr.assign_coords({'time': time1})
                arr = arr.expand_dims('time')

        if not 'height' in dims:
            if not isinstance(height, (int, float)):
                raise ValueError("If height is not part of the array's dimensions, then a int or float must be passed to height.")
            else:
                print('height will be added to the data.')
                arr = arr.assign_coords({'height': height})
                arr = arr.expand_dims('height')

        setattr(self, 'data', arr)
        setattr(self, 'parameter', parameter)

        pass


    def determine_grid_block_size(self, starting_x_size=10, starting_y_size=10, increment=10, min_size=800, max_size=1100):
        """
        Function to determine the appropriate grid size for splitting.

        Parameters
        ----------
        arr : DataArray
            An xarray DataArray with at least x and y dimensions. It can have any number of dimensions, though it probably does not make much sense to have greater than 4 dimensions.
        starting_x_size : int
            The initial size or length of the smaller grids in the x dimension.
        starting_y_size : int
            The initial size or length of the smaller grids in the y dimension.
        increment : int
            The incremental grid size to be added iteratively to the starting sizes.
        min_size : int
            The minimum acceptable object size in KB.
        max_size : int
            The maximum acceptable object size in KB.
        x_name : str
            The x dimension name.
        y_name : str
            The y dimension name.

        Returns
        -------
        dict
            Of the optimised grid size results.
        """
        encoding = {'lon': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 1e-07},
 'lat': {'dtype': 'int32', '_FillValue': -999999, 'scale_factor': 1e-07},
 'altitude': {'dtype': 'int32', '_FillValue': -9999, 'scale_factor': 0.001},
 'time': {'_FillValue': -99999999, 'units': 'days since 1970-01-01 00:00:00'}}

        encoding.update(self.datasets[0]['properties']['encoding'])

        max_obj_size = 0
        x_size = starting_x_size
        y_size = starting_y_size

        while True:
            block_list = _split_grid(self.data[self.parameter], x_size=x_size, y_size=y_size)
            block_list2 = [a.to_dataset() for a in block_list]

            for block in block_list2:
                for k, v in encoding.items():
                    if k in block:
                        block[k].encoding = v

            obj_sizes = [len(write_pkl_zstd(nc.to_netcdf())) for nc in block_list2]
            max_obj_size = max(obj_sizes)

            if max_obj_size < min_size*1000:
                x_size = x_size + increment
                y_size = y_size + increment
            else:
                break

        if max_obj_size > max_size*1000:
            print('max_object_size:', str(max_obj_size))
            raise ValueError('max object size is greater than the allotted size. Reduce the increment value and start again.')

        obj_dict = {'x_size': x_size, 'y_size': y_size, 'max_obj_size': max_obj_size, 'min_obj_size': min(obj_sizes), 'sum_obj_size': sum(obj_sizes), 'len_obj': len(obj_sizes)}

        setattr(self, 'grid_size_dict', obj_dict)

        return obj_dict


    def save_results(self, x_size=None, y_size=None, threads=30):
        """

        """
        ## Create the data_dict
        data_dict = {d['dataset_id']: [] for d in self.datasets}

        ## Process data
        print('Processing data...')

        if self.grouping == 'blocks':
            if not (isinstance(x_size, int) and isinstance(y_size, int)):
                raise TypeError('x_size and y_size must be ints.')

            block_list = _split_grid(self.data[self.parameter], x_size=x_size, y_size=y_size)
            # TODO: add in option to save ancilliary variables

            for block in block_list:
                prepare_results(data_dict, self.datasets, block.to_dataset(), self.max_run_date_key)
        else:
            data1 = self.data.stack(geometry=['lon', 'lat'])
            g2 = data1.groupby('geometry', squeeze=False)

            for index, g in g2:
                res1 = g.unstack('geometry')
                prepare_results(data_dict, self.datasets, res1, self.max_run_date_key)

        ## Update to S3
        print('Updating data on S3')
        update_results_s3(self.processing_code, data_dict, self.run_date_dict, self.s3_remote, threads, self.public_url)

        print('Processing and saving data has been successful!')


    def update_aggregates(self, threads=60):
        """

        """
        ## Update the datasets and station jsons
        print('Aggregating dataset and station data.')
        s3 = s3_connection(self.s3_remote['connection_config'], threads)

        for ds in self.datasets:
            ds_new = put_remote_dataset(s3, self.s3_remote['bucket'], ds)
            ds_stations = put_remote_agg_stations(s3, self.s3_remote['bucket'], ds['dataset_id'], threads)

        # Aggregate all datasets for the bucket
        ds_all = put_remote_agg_datasets(s3, self.s3_remote['bucket'], threads)

        print('Updating the aggregates has been successful!')




















"""
Functions to query APIs for altitude data.

"""
import requests
import pandas as pd
import numpy as np
from tethysts import Tethys
from shapely.geometry import mapping
from shapely import wkb

######################################
#### Parameters

koordinates_rater_format = '{base_url}/services/query/v1/raster.json?key={key}&layer={layer_id}&x={lon}&y={lat}'

dem_remote = {'bucket': 'nasa-data', 'connection_config': 'https://b2.tethys-ts.xyz'}
dem_ds_id = '4690821a39b3fb65c197a540'

#####################################
### Functions


def koordinates_raster_query(base_url: str, key: str, layer_id: (int, str), lon: float, lat: float):
    """

    """
    url_request = koordinates_rater_format.format(base_url=base_url, key=key, layer_id=layer_id, lon=lon, lat=lat)
    resp = requests.get(url_request)

    if resp.ok:
        layer1 = resp.json()['rasterQuery']['layers'][str(layer_id)]

        status = layer1['status']

        if status == 'ok':
            bands = layer1['bands']

            return bands
        else:
            print('status is: ' + status)
            return None

    else:
        return resp.content.decode()


def get_altitude(stn_df, dem_remote=dem_remote, dem_ds_id=dem_ds_id, source_ds_id=None, source_remote=None):
    """

    """
    stn_df1 = stn_df[['station_id', 'geometry']].drop_duplicates(subset=['station_id']).copy()

    ## Get existing stns
    if isinstance(source_remote, dict) and isinstance(source_ds_id, str):
        try:
            t1 = Tethys([source_remote])
            old_stns = t1.get_stations(source_ds_id)
        except:
            old_stns = []
    else:
        old_stns = []

    ## Combine new stations with existing stations
    if old_stns:
        old_stns_alt = [{'station_id': s['station_id'], 'altitude': s['altitude']} for s in old_stns if 'altitude' in s]
        if old_stns_alt:
            old_stns_df = pd.DataFrame(old_stns_alt)
            combo1 = pd.merge(stn_df1, old_stns_df, how='left', on='station_id')
        else:
            combo1 = stn_df1.copy()
            combo1['altitude'] = np.nan
    else:
        combo1 = stn_df1.copy()
        combo1['altitude'] = np.nan

    ## Iterate through stations for altitude
    t1 = Tethys([dem_remote])

    alt_list = []
    for i, row in combo1.iterrows():
        # print(i)
        if np.isnan(row['altitude']):
            geo1 = mapping(wkb.loads(row['geometry'], True))
            results = t1.get_results(dem_ds_id, lon=float(geo1['coordinates'][0]), lat=float(geo1['coordinates'][1]), squeeze_dims=True, output='DataArray', cache='memory')
            alt = int(results)
        else:
            alt = int(row['altitude'])

        alt_list.append({'station_id': row['station_id'], 'altitude': alt})

    ## Package up and return
    alt_df = pd.DataFrame(alt_list)

    return alt_df


####################################
### Testing

# base_url = 'https://data.linz.govt.nz'
# key = ''
# layer_id = 51768
# lon = 172.084790
# lat = -43.222752
#
#
#
# bands1 = koordinates_raster_query(base_url, key, layer_id, lon, lat)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 20 10:43:15 2021

@author: mike
"""
import requests
import orjson
import copy

##########################################
### Functions


def query_esri_mapserver(base_url, layer_id, out_fields=None, where=None):
    """
    Query an ESRI map server for a vector layer and return a geojson structured dict.

    Parameters
    ----------
    base_url : str
        The base url for the map server up to the layer_id.
    layer_id : str or int
        The layer id.
    out_fields : None or list of str
        The output fields to be returned. The geometry will always be returned. None will return all fields.
    where : None or str
        The SQL style 'where' clause to query the layer. None will have no filters.

    Returns
    -------
    Dict
        In GeoJSON structure
    """
    url = base_url + str(layer_id) + '/query'

    ## Set up filters
    params = {'returnGeometry': 'true', 'f': 'geojson'}

    if out_fields is None:
        params.update({'outFields': '*'})
    elif isinstance(out_fields, list):
        out_str = ', '.join(out_fields)
        params.update({'outFields': out_str})
    else:
        raise ValueError('out_fields must either be None or a list of strings')

    if where is None:
        params.update({'where': '1=1'})
    elif isinstance(where, str):
        params.update({'where': where})
    else:
        raise ValueError('where must either be None or a string')

    ## Run the queries
    resp = requests.get(url, params=params)

    if not resp.ok:
        raise ValueError(resp.raise_for_status())

    data = orjson.loads(resp.content)

    if 'exceededTransferLimit' in data:
        exceed = data.pop('exceededTransferLimit')
    else:
        exceed = False

    data_all = copy.deepcopy(data)

    while exceed:
        last_id = data['features'][-1]['id']

        if isinstance(where, str):
            params.update({'where': where + 'and objectid > ' + str(last_id)})
        else:
            params.update({'where': 'objectid > ' + str(last_id)})

        resp = requests.get(url, params=params)

        if not resp.ok:
            raise ValueError(resp.raise_for_status())

        data = orjson.loads(resp.content)

        data_all['features'].extend(data['features'])

        if 'exceededTransferLimit' in data:
            exceed = data.pop('exceededTransferLimit')
        else:
            exceed = False

    return data_all


































# -*- coding: utf-8 -*-
"""
Created on Tue Jul  2 09:25:41 2019

@author: michaelek
"""
import os
import yaml
# import pytest
import pandas as pd

pd.options.display.max_columns = 10


###############################################
### Parameters

base_dir = os.path.realpath(os.path.dirname(__file__))

with open(os.path.join(base_dir, 'parameters.yml')) as param:
    param = yaml.safe_load(param)

source = param['source'].copy()
remote = param['remote']
public_url = source['public_url']
processing_code = source['processing_code']

datasets = source['datasets']
parameter = datasets[0]['parameter']

########################################
### Tests


self = Grid(datasets, remote, processing_code, public_url)

self.load_data(data, parameter, height=2)

block_test = self.determine_grid_block_size(starting_x_size=20, starting_y_size=20, increment=10, min_size=800, max_size=1100)

x_size = block_test['x_size']
y_size = block_test['y_size']



































